import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;

import Jama.Matrix;

import edu.umbc.cs.maple.utils.JamaUtils;
import edu.umbc.cs.maple.utils.JavaUtils;
import edu.umbc.cs.maple.utils.MathUtils;
import edu.umbc.cs.maple.utils.SGTUtils;

/** Interactive Community Detection
 * <p>
 * Copyright (c) 2011 Eric Eaton, Rachael Mansbach
 * <p>
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 * 
 * @author Eric Eaton (eeaton@cs.brynmawr.edu), Bryn Mawr College
 * @author Rachael Mansbach (rmansba@swarthmore.edu), Swarthmore College
 * 
 * @version 1.0
 *
 */

public class GuidedCommunityDetection {
	
	public static double m_objFnValue = 0;
	
	public static double[][] propagateLabels(Jama.Matrix adjacencyMatrix, double[][] functionalLabels) {
		Jama.Matrix graphLaplacian = SGTUtils.undirectedWeightedAdjacencyToLaplacian(adjacencyMatrix, SGTUtils.LaplacianType.NORMALIZED);
		return propagateLabels(graphLaplacian.eig(), functionalLabels);
	}
	
	
	public static double[][] propagateLabels(Jama.EigenvalueDecomposition graphLaplacianEVD, double[][] functionalLabels) {
		
		Jama.Matrix mQ = graphLaplacianEVD.getV();
		Jama.Matrix mLambda = graphLaplacianEVD.getD();
		
		// set the eigenvectors to be constant for all eigenvalues of 0
		double[] ones = new double[mQ.getRowDimension()];
		for (int i=0; i<ones.length; i++) ones[i] = 1;
		double[] eigenvalues = JamaUtils.getDiagonal(mLambda);
		for (int i=0; i<eigenvalues.length; i++) {
			if (Math.abs(eigenvalues[i]) < 1E-12)  // i.e., eigenvalue is approx. 0
				JamaUtils.setcol(mQ, i, ones);
		}
		
		// compute the sqrt of lambda for more regularization
		int numPoints = mLambda.getRowDimension();
		Jama.Matrix mRegOperatorLambda = mLambda.copy();
		for (int i=0; i<numPoints; i++) {
			double value = Math.sqrt(mRegOperatorLambda.get(i,i));
			if (!Double.isNaN(value))
				mRegOperatorLambda.set(i,i,Math.max(value,1E-10));
			else
				mRegOperatorLambda.set(i,i,1E-10);
		}
		double[] mRegOperatorLambdaDiag = JamaUtils.getDiagonal(mRegOperatorLambda);
		
		int numVertices = mQ.getRowDimension();
		
		// determine the labeled vertices and the length of the labels
		int labelLen = 0;
		int numLabels = 0;
		for (int i=0; i<functionalLabels.length; i++) {
			double[] label = functionalLabels[i];
			if (label != null) {
				labelLen = Math.max(label.length, labelLen);
				numLabels++;
			}
		}
		
		if (numLabels == 0)
			return functionalLabels;
			
		// convert the training data's labels to a matrix
		Jama.Matrix f = new Jama.Matrix(numLabels,labelLen);
		int fRowIdx = 0;
		int[] labeledIndices = new int[numLabels];
		for (int i=0; i<functionalLabels.length; i++) {
			double[] label = functionalLabels[i];
			if (label != null) {
				JamaUtils.setrow(f,fRowIdx,label);
				labeledIndices[fRowIdx] = i;
				fRowIdx++;
			}
		}
		
		if (numLabels == 1) {
			double[] value = f.getColumnPackedCopy();
			double[][] propagatedLabels = new double[numVertices][];
			for (int i=0; i < numVertices; i++) {
				propagatedLabels[i] = JavaUtils.arrayCopy(value);
			}
			return propagatedLabels;
		}
		
		// get the set of corresponding eigenvectors and eigenvalues
		Jama.Matrix q = JamaUtils.getrows(mQ, labeledIndices);	
		Jama.Matrix l = mRegOperatorLambda;
				
		// solve f = qw, regularized by l
		Jama.Matrix w = JamaUtils.regLeastSquares(q, f, l);
	
		// recompute the current data
		Jama.Matrix newF = mQ.times(w);


		if (true) {
			System.out.println("RMSE to original= "+JamaUtils.rmse(JamaUtils.getrows(newF,labeledIndices),f));
		}
		
		// store the values for the data
		double[][] propagatedLabels = new double[numVertices][];
		for (int i=0; i < numVertices; i++) {
			// store the original label if given
			if (functionalLabels[i] != null) {
				propagatedLabels[i] = JavaUtils.arrayCopy(functionalLabels[i]);
			} else {
				propagatedLabels[i] = JavaUtils.arrayCopy(JamaUtils.getrow(newF, i).getColumnPackedCopy());
			}
		}
		
		return propagatedLabels;
	}
	
	
	public static int[] findCommunities(Matrix adjacencyMatrix, List<Constraint> constraints,
			int numCommunities, boolean lockNumCommunities, double mu) {
		
		// precompute aspects of the graph
		double[] vertexDegrees = JamaUtils.colsum(adjacencyMatrix).getRowPackedCopy();
		int m = (int) MathUtils.sum(vertexDegrees) / 2;  // assuming an undirected graph
		
		HashMap<String,Double> mustlinkConstraints = new HashMap<String,Double>();
		HashMap<String,Double> cannotlinkConstraints = new HashMap<String,Double>();
		
		// load the constraints
		for (Constraint c : constraints) {
			if (c.getConstraintType() == Constraint.ConstraintType.MUST_LINK) {
				mustlinkConstraints.put(c.getIdxDescription(), c.getWeight());
			} else if (c.getConstraintType() == Constraint.ConstraintType.CANNOT_LINK) {
				cannotlinkConstraints.put(c.getIdxDescription(), c.getWeight());
			} else {
				System.out.println("Warning:  constraint type is unspecified: <"+c.toString()+">");
			}
		}
		
		// compute the modularity and userGuidance matrices
		Matrix modularityMatrix = computeModularityMatrix(adjacencyMatrix, vertexDegrees, m);
		Matrix userGuidanceMatrix = computeUserGuidanceMatrix(adjacencyMatrix, mustlinkConstraints, cannotlinkConstraints);
		
		return findCommunities(modularityMatrix, userGuidanceMatrix, numCommunities, lockNumCommunities, mu);
	}
	
	
	public static int[] findCommunities(Matrix adjacencyMatrix, double[][] functionalLabels,
			int numCommunities, boolean lockNumCommunities, double mu) {
		
		// precompute aspects of the graph
		double[] vertexDegrees = JamaUtils.colsum(adjacencyMatrix).getRowPackedCopy();
		int m = (int) MathUtils.sum(vertexDegrees) / 2;  // assuming an undirected graph
		
		// compute the modularity and userGuidance matrices
		Matrix modularityMatrix = computeModularityMatrix(adjacencyMatrix, vertexDegrees, m);
		Matrix userGuidanceMatrix = computeUserGuidanceMatrix(adjacencyMatrix, vertexDegrees, m, functionalLabels);
		
		return findCommunities(modularityMatrix, userGuidanceMatrix, numCommunities, lockNumCommunities, mu);
	}
	
	
	/** Uses simulated annealing to find the communities given the undirected adjacency matrix and functional labels
	 * @param adjacencyMatrix
	 * @param functionalLabels
	 * @return
	 */
	public static int[] findCommunities(Matrix modularityMatrix, Matrix userGuidanceMatrix, int numCommunities, boolean lockNumCommunities, double mu) {
		
		int numVertices = modularityMatrix.getRowDimension();
		
		// initialize the simulated annealing parameters and result tracking
		int maxSimAnnealingIterations = 10;
		double initialTemp = 10000;
		double tempDecreaseRate = 0.985;
		Random randGenerator = MathUtils.getRandomGenerator();
		
		// randomly initialize the community structure and corresponding obj fn
		int[] bestCommunitiesSoFar = new int[numVertices];
		for (int i=0; i < numVertices; i++) {
			bestCommunitiesSoFar[i] = randGenerator.nextInt(numCommunities);
		}
		bestCommunitiesSoFar = compressCommunities(bestCommunitiesSoFar);
		double bestObjFnValueSoFar = computeUserGuidedModularity(modularityMatrix, userGuidanceMatrix, bestCommunitiesSoFar, mu);
		
		for (int simAnnealingIterationNum = 0; simAnnealingIterationNum < maxSimAnnealingIterations; simAnnealingIterationNum++) {
			
			// start where we left off previously
			int[] communities = bestCommunitiesSoFar;
			double objFnValue = bestObjFnValueSoFar;
			
			//System.out.println("Running simulated annealing (iter "+simAnnealingIterationNum+" of "+maxSimAnnealingIterations+")");
			
			// run simulated annealing
			int step = 0;
			double temp = initialTemp;
			while (temp > 1E-8) {
				
				//if (step % 50 == 0)
					//System.out.println("  Step="+step+"\tTemp="+JavaUtils.format4DecimalPlaces.format(temp)+ "\tObjFnValue="+JavaUtils.format4DecimalPlaces.format(objFnValue));
				
				// choose a random vertex and change its community
				int vertexToChange = randGenerator.nextInt(numVertices);
				
				int numCommunitiesToSample = -1;
				if (lockNumCommunities) {
					numCommunitiesToSample = numCommunities;
				} else {
					// compute the number of communities
					int currentNumCommunities = MathUtils.maxValue(communities)+1;
					numCommunitiesToSample = currentNumCommunities+1;
				}
				int newCommunity = randGenerator.nextInt(numCommunitiesToSample);
				if (newCommunity == communities[vertexToChange]) {
					newCommunity = (newCommunity+1) % numCommunitiesToSample;
				}
				int[] newCommunities = JavaUtils.arrayCopy(communities);
				newCommunities[vertexToChange] = newCommunity;
				
				// compute the new objective function
				double newObjFnValue = computeUserGuidedModularity(modularityMatrix, userGuidanceMatrix, newCommunities, mu);
				
				// test whether we should move to it
				if (newObjFnValue < objFnValue) {
					// move to the new community state
					if (lockNumCommunities)
						communities = newCommunities;  // no need to compress the communities if the number is fixed
					else
						communities = compressCommunities(newCommunities);
					objFnValue = newObjFnValue;
				}
				// otherwise, move to it with some probability
				double probOfMoving = Math.exp(-(newObjFnValue - objFnValue) / temp);
				if (randGenerator.nextDouble() < probOfMoving) {
					// move to the new community state
					if (lockNumCommunities)
						communities = newCommunities;  // no need to compress the communities if the number is fixed
					else
						communities = compressCommunities(newCommunities);
					objFnValue = newObjFnValue;
				}
				// increment the step count and decrease temperature
				temp = temp * tempDecreaseRate;
				step++;
			}
			
			if (objFnValue < bestObjFnValueSoFar) {
				bestObjFnValueSoFar = objFnValue;
				bestCommunitiesSoFar = communities;
			}
		}
		
		
		System.out.println("  Final ObjFnValue="+JavaUtils.format4DecimalPlaces.format(bestObjFnValueSoFar));
		
		m_objFnValue = bestObjFnValueSoFar;
		
		return bestCommunitiesSoFar;
	}
	
	/** Takes a set of communities numbered randomly and renumbers them in increasing order
	 * 
	 * @param c
	 * @return
	 */
	static int[] compressCommunities(int[] communities) {
		// get the set of unique numbers
		HashSet<Integer> uniques = new HashSet<Integer>();
		for (int c : communities)
			uniques.add(c);
		// determine the renumbering, starting with zero
		HashMap<Integer,Integer> renumbering = new HashMap<Integer,Integer>();
		int count = 0;
		for (int c : uniques) {
			renumbering.put(c,count);
			count++;
		}
		// renumber the specified communities
		int[] newCommunities = new int[communities.length];
		for (int i=0; i<newCommunities.length; i++) {
			newCommunities[i] = renumbering.get(communities[i]);
		}
		return newCommunities;
	}
	
	/** Compute the modularity matrix (A - P)
	 * @param adjacencyMatrix
	 * @param vertexDegrees
	 * @param m
	 * @return the modularity matrix
	 */
	public static Matrix computeModularityMatrix(Matrix adjacencyMatrix, double[] vertexDegrees, int m) {
		
		double twoM = 2 * m;
		
		int numRows = adjacencyMatrix.getRowDimension();
		int numCols = adjacencyMatrix.getColumnDimension();
		
		Matrix modularityMatrix = JamaUtils.zeros(numRows, numCols);
		
		for (int i=0; i < numRows; i++) {
			double d_i = vertexDegrees[i];
			
			for (int j=0; j < numCols; j++) {
				
				// compute the variables of the equation
				double A_ij = adjacencyMatrix.get(i,j);
				double d_j = vertexDegrees[j];
				
				// compute the null model
				double nullProbability = d_i*d_j / twoM;
				
				// compute the modularity
				double M_ij =  A_ij - nullProbability;
				modularityMatrix.set(i,j,M_ij);
			}
		}
		
		return modularityMatrix;
	}
	
	
	/** Compute the guided modularity Hamiltonian
	 *  -\sum_{i \neq j} M_{ij} \delta(C_i,C_j) - \mu \sum_{i \neq j} \Delta U_{ij} \delta(C_i,C_j)
	 * @param modularityMatrix
	 * @param userGuidanceMatrix
	 * @param communities
	 * @param mu
	 * @return
	 */
	public static double computeUserGuidedModularity(Matrix modularityMatrix, Matrix userGuidanceMatrix, 
			int[] communities, double mu) {
		
		double sum = 0;
		
		int numRows = modularityMatrix.getRowDimension();
		int numCols = modularityMatrix.getColumnDimension();
		
		for (int i=0; i < numRows; i++) {
			for (int j=0; j < numCols; j++) {
				
				if (i == j) continue;
				
				// only contribute to the value if delta(C_i, C_j) == 1
				if (communities[i] != communities[j])
					continue;
				
				double M_ij = modularityMatrix.get(i,j);
				double deltaU_ij = userGuidanceMatrix.get(i,j);
				
				sum += M_ij + mu * deltaU_ij;    // we're adding instead of subtracting, and will negate later
			}
		}
		
		return (-sum);   // negate the sum, since we're adding instead of subtracting 
	}
	
	
	
	/** Compute the user guidance matrix
	 * @param adjacencyMatrix
	 * @param mustlinkConstraints the weights of the must-link constraints, keyed by Constraint.getIdxDescription()
	 * @param cannotlinkConstraints the weights of the cannot-link constraints, keyed by Constraint.getIdxDescription()
	 * @return the user guidance matrix
	 */
	public static Matrix computeUserGuidanceMatrix(Matrix adjacencyMatrix, 
			HashMap<String,Double> mustlinkConstraints, HashMap<String,Double> cannotlinkConstraints) {
		
		int numRows = adjacencyMatrix.getRowDimension();
		int numCols = adjacencyMatrix.getColumnDimension();
		
		Matrix userGuidanceMatrix = JamaUtils.zeros(numRows, numCols);
		
		for (int i=0; i < numRows; i++) {
			for (int j=0; j < numCols; j++) {
				
				// set alpha1 and alpha2 to be 1 for now
				double alpha1 = 1;
				double alpha2 = 1;
				
				String idxDescription = (new Constraint(i,j)).getIdxDescription();
				
				double w_ij =  mustlinkConstraints.containsKey(idxDescription) ? mustlinkConstraints.get(idxDescription) : 0;
				double wbar_ij =  cannotlinkConstraints.containsKey(idxDescription) ? cannotlinkConstraints.get(idxDescription) : 0;
				
				double u_ij = alpha1 * w_ij;
				double ubar_ij = alpha2 * wbar_ij;
				
				// compute deltaU_ij = u_ij - ubar_ij
				double deltaU_ij = u_ij - ubar_ij;
				
				userGuidanceMatrix.set(i,j,deltaU_ij);
			}
		}
		
		return userGuidanceMatrix;
	}
	
	
	
	/** Compute the user guidance matrix
	 * @param adjacencyMatrix
	 * @param vertexDegrees
	 * @param m
	 * @param functionalLabels
	 * @return the user guidance matrix
	 */
	public static Matrix computeUserGuidanceMatrix(Matrix adjacencyMatrix, 
			double[] vertexDegrees, int m, double[][] functionalLabels) {
		
		int numRows = adjacencyMatrix.getRowDimension();
		int numCols = adjacencyMatrix.getColumnDimension();
		
		Matrix userGuidanceMatrix = JamaUtils.zeros(numRows, numCols);
		
		for (int i=0; i < numRows; i++) {
			double d_i = vertexDegrees[i];
			
			for (int j=0; j < numCols; j++) {
				
				// set alpha1 and alpha2 to be 1 for now
				double alpha1 = 1;
				double alpha2 = 1;
				
				// compute the variables of the equation
				double A_ij = adjacencyMatrix.get(i,j);
				double d_j = vertexDegrees[j];
				double[] f_i = functionalLabels[i];
				double[] f_j = functionalLabels[j];
				
				// if not guidance is available, the entry is zero
				if (f_i == null || f_j == null) {
					userGuidanceMatrix.set(i,j,0);
					continue;
				}
				
				double u_ij = alpha1 * computeSmoothness(A_ij, f_i, f_j, d_i, d_j);
				double ubar_ij = -alpha2 * computeSmoothness(A_ij, f_i, f_j, d_i, d_j);
				
				// compute deltaU_ij = u_ij - ubar_ij
				double deltaU_ij = u_ij - ubar_ij;
				
				userGuidanceMatrix.set(i,j,deltaU_ij);
			}
		}
		
		return userGuidanceMatrix;
	}
	
	
	
	/** Compute the modularity matrix with user guidance.
	 * @param adjacencyMatrix
	 * @param vertexDegrees
	 * @param m
	 * @param mu
	 * @param functionalLabels
	 * @return the modularity matrix
	 */
	public static Matrix computeGuidedModularityMatrix(Matrix adjacencyMatrix, 
			double[] vertexDegrees, int m, double mu, double[][] functionalLabels) {
		

		double twoM = 2 * m;
		
		int numRows = adjacencyMatrix.getRowDimension();
		int numCols = adjacencyMatrix.getColumnDimension();
		
		Matrix modularityMatrix = JamaUtils.zeros(numRows, numCols);
		
		for (int i=0; i < numRows; i++) {
			double d_i = vertexDegrees[i];
			
			for (int j=0; j < numCols; j++) {
				
				// set alpha1 and alpha2 to be 1 for now
				double alpha1 = 1;
				double alpha2 = 1;
				
				// compute the variables of the equation
				double A_ij = adjacencyMatrix.get(i,j);
				double d_j = vertexDegrees[j];
				double[] f_i = functionalLabels[i];
				double[] f_j = functionalLabels[j];
				double u_ij = alpha1 * computeSmoothness(A_ij, f_i, f_j, d_i, d_j);
				double ubar_ij = -alpha2 * computeSmoothness(A_ij, f_i, f_j, d_i, d_j);
				
				if (Double.isNaN(u_ij) || Double.isNaN(ubar_ij))
					throw new IllegalStateException("u_ij or ubar_ij is NaN:  A_ij="+A_ij+" d_i="+d_i+" d_j"+d_j+" f_i="+JavaUtils.arrayToString(f_i)+" f_j="+JavaUtils.arrayToString(f_j));
				
				// compute the null model
				double nullProbability = d_i*d_j / twoM;
				
				// compute the guidance portion
				double guidanceModifier = u_ij - ubar_ij;
				
				// compute the modularity
				double M_ij =  A_ij - (nullProbability - mu * guidanceModifier);
				modularityMatrix.set(i,j,M_ij);
			}
		}
		
		return modularityMatrix;
	}
	
	private static double computeSmoothness(double A_ij, double[] f_i, double[] f_j, double d_i, double d_j) {
		double[] normalizedFi = null;
		if (d_i == 0) {
			normalizedFi = MathUtils.vectorMultiply(0,f_i);
		} else {
			normalizedFi = MathUtils.arrayDivide(f_i, Math.sqrt(d_i));
		}
		double[] normalizedFj = null;
		if (d_j == 0) {
			normalizedFj = MathUtils.vectorMultiply(0, f_j);
		} else {
			normalizedFj = MathUtils.arrayDivide(f_j, Math.sqrt(d_j));
		}
		double[] diffFiFj = MathUtils.vectorSubtract(normalizedFi, normalizedFj);
		return A_ij * MathUtils.norm2Squared(diffFiFj);
	}
	

}
